/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef MAPS_HPP__
#define MAPS_HPP__

#include <boost/algorithm/string/predicate.hpp>
#include <boost/utility/string_ref.hpp>
#include <fstream>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

enum class Permissions {
	None  = 0,
	Exec  = 1 << 0,
	Write = 1 << 1,
	Read  = 1 << 2,
	Priv  = 1 << 3
};

inline Permissions operator|(Permissions a, Permissions b) {
	return static_cast<Permissions>(static_cast<int>(a) | static_cast<int>(b));
}

inline Permissions operator&(Permissions a, Permissions b) {
	return static_cast<Permissions>(static_cast<int>(a) & static_cast<int>(b));
}

struct MapRecord {
	uintptr_t start;
	uintptr_t end;
	uintptr_t offset;
	Permissions perms;
	std::string path;
	bool important;

	bool IsLoadable() const {
		// we need only readable records
		bool result = (perms & Permissions::Read) == Permissions::Read;

		// we don't need [vvar] becasue there are problems with reading
		// on x86, and [vvar] doesn't contain useful data to stack resolving
		result &= path != "[vvar]";

		// we don't need executable, unless it is [vsyscall] or [vdso]
		// https://lwn.net/Articles/615809/
		result &= ((perms & Permissions::Exec) != Permissions::Exec) ||
			(path == "[vsyscall]" || path == "[vdso]");

		result &= !boost::algorithm::ends_with(path, "locale-archive");

		return result;
	}
};

class Maps {
 private:
	const pid_t m_pid;
	std::vector<std::shared_ptr<MapRecord>> m_map_records;

 public:
	explicit Maps(const pid_t pid) : m_pid(pid) {}

	const std::vector<std::shared_ptr<MapRecord>> &Records() const {
		return m_map_records;
	}

	bool DoWeNeedThisRecord(const std::shared_ptr<MapRecord> record) {
		return (record->path == "[vsyscall]" ||
		        record->path == "[vdso]"     ||
		        record->path == "[vectors]"  ||
		        record->path == "[stack]");
	}

	void Parse() {
		const std::string& maps_file_name = "/proc/" + std::to_string(m_pid) + "/maps";

		std::ifstream maps_file(maps_file_name);
		if (!maps_file)
			throw std::system_error(errno, std::system_category(), "failed to open mem file " + maps_file_name);

		std::string line;
		while (std::getline(maps_file, line)) {
			std::shared_ptr<MapRecord> record = ParseLine(line);

			record->important = DoWeNeedThisRecord(record);
			if ((record->perms & Permissions::Read) == Permissions::Read)
				m_map_records.push_back(record);
		}
	}

	// Parse line of /proc/<pid>/maps file that has the format specifier:
	//
	//     "%08lx-%08lx %c%c%c%c %08llx %02x:%02x %lu "
	//
	std::shared_ptr<MapRecord> ParseLine(boost::string_ref line) {
		static const char FIELD_DELIMITER = ' ';
		std::shared_ptr<MapRecord> record = std::make_shared<MapRecord>();

		auto getNextToken = [&line](const char delim) -> boost::string_ref {
			auto pos = line.find_first_of(delim);
			auto result = line.substr(0, pos);
			line = line.substr(pos + 1);
			return result;
		};

		// START
		record->start = std::stoull(std::string(getNextToken('-')), nullptr, 16);

		// END
		record->end = std::stoull(std::string(getNextToken(FIELD_DELIMITER)), nullptr, 16);

		// PERMS
		auto token = getNextToken(FIELD_DELIMITER);
		record->perms = Permissions::None;

		auto storePerm = [&record, &token](const char c, const Permissions perm) {
			if (token.find(c) != std::string::npos)
				record->perms = record->perms | perm;
		};

		storePerm('r', Permissions::Read);
		storePerm('w', Permissions::Write);
		storePerm('x', Permissions::Exec);
		storePerm('p', Permissions::Priv);

		// OFFSET
		record->offset = std::stoull(std::string(getNextToken(FIELD_DELIMITER)), nullptr, 16);

		// DEV - ignore
		getNextToken(FIELD_DELIMITER);

		// INODE - ignore
		getNextToken(FIELD_DELIMITER);

		// PATH
		size_t delim = line.find_first_not_of(' ');
		if (delim != std::string::npos) {
			line = line.substr(delim);
			record->path = std::string(line);
		} else {
			record->path.clear();
		}
		return record;
	}
};

#endif  // MAPS_HPP__
